#include <stdio.h>
#include <ctype.h>
#include <string.h>  // required for strchrnul()
#include <stdlib.h>
#include <stdint.h>

uint8_t sum;
uint8_t *inbf, *outbf;
uint32_t filesize;

int terminal_error(void){
  if(inbf != NULL)
    free(inbf);

  if(outbf != NULL)
    free(outbf);

  exit(-1);
}

uint32_t file_save(const char * filename, void * buf, uint32_t len) {
  FILE * out = fopen(filename, "wb");
  int r;
  if(out == NULL)
    return 0;

  r = fwrite(buf, len, 1, out);
  fclose(out);
  return r;
}

void * file_load(const char * filename, uint32_t * len) {
  uint32_t size, n;
  FILE * in;
  char * buf;
  in = fopen(filename, "rb");

  fseek(in, 0, SEEK_END);       // seek to end of file
  size = ftell(in);             // get current file pointer
  fseek(in, 0, SEEK_SET);       // seek back to beginning of file
  buf = malloc(size);        // allocate size
  if(buf == NULL){
    printf("Unable to allocate file buffer!\r\n");
    return NULL;
  }
  n = fread(buf, 1, size, in);  // Read whole file
  if(n < size){                 // Ensure it was completely read
    free(buf);
    return NULL;
  }

  if(len)
    *len = n;

  if(in != stdin)
    fclose(in);

  return buf;
}

void * _memmem(const void *haystack, size_t haystack_len, const void * const needle, const size_t needle_len) {
  if (haystack == NULL) return NULL; // or assert(haystack != NULL);
  if (haystack_len == 0) return NULL;
  if (needle == NULL) return NULL; // or assert(needle != NULL);
  if (needle_len == 0) return NULL;

  for (const char *h = haystack; haystack_len >= needle_len; ++h, --haystack_len) {
    if (!memcmp(h, needle, needle_len)) {
      return (char *)h;
    }
  }
  return NULL;
}

void convertToUpperCase(char * s){
  if(s[0] && s[1]=='\0'){                                   // Fix itoa output, as it doesn't fill single-digit numbers with 0 (Ex. makes "B" instead "0B")
    s[1] = s[0];
    s[0] = '0';
  }
  while(*s){
    if( (*s>='a') && (*s<='f') ){
      *s -= 32;
    }
    s++;
  }
}

void intelHexToArray(char *in, uint8_t *dout){              // Intel hex string to uint8_t[16] array
  if(in == NULL){
    printf("No valid data found!\r\n");
    terminal_error();
  }
  if(in == NULL){
    printf("Invalid dst pointer!\r\n");
    terminal_error();
  }

  uint8_t * b = in;
  uint8_t checksum=0, xor_byte=0, chk_byte=0;
  char s[3] = {0};
                                                            // Intel hex structure used
                                                            // :10      0000        00      00000000000000000000000000000000    00
  for(uint8_t i=0; i<21; i++){                              //  0(size) 1-2(addr)   3(xor)  4(Data)                             20(chksum)
    s[0]=*b++;                                              // Copy 2 chars
    s[1]=*b++;
    uint8_t data = strtol((const char*)s, NULL, 16);        // Convert to integer
    if(i<20){
     checksum += data;                                      // Perform checksum of bytes 0-19
    }
    if(i==3){
      xor_byte = data;                                      // Byte 3 is xor value we need to decrypt the data
    }
    else if(i>=4 && i<20){                                  // Bytes 4-19 are the fw data, decrypt with xor
      *dout++ = data ^ (xor_byte + 0x9A);
    }
    else if(i==20){
      chk_byte = data;                                      // Byte 20 is the checksum, store to compare with the computed checksum
    }
  }
  checksum ^= 0xFF;                                         // Perfom 2's complement on the computed checksum
  checksum += 1;
  if(checksum != chk_byte){                                 // Both must must match, otherwise data is corrupted
    printf("Bad checksum!\r\n");
    printf("XOR:%x  Checksum:%x  Computed:%x\r\n",xor_byte,chk_byte,checksum);
    char t [50];
    memcpy(t,b,42);
    t[42]=0;
    printf("Data: %s\r\n",t);
    terminal_error();
  }
}

void arrayToIntelHex(uint8_t *din, char * out){            // uint8_t[16] array to Intel hex string
    if(din == NULL){
        printf("No valid data found!\r\n");
        terminal_error();
    }
    if(out == NULL){
        printf("Invalid dst pointer!\r\n");
        terminal_error();
    }

    uint8_t * b = out;
    uint8_t checksum=0,xor_byte=0;
    char s[3] = {0};
                                                            // Intel hex structure used
                                                            // :10      0000        00      00000000000000000000000000000000    00
    for(uint8_t i=0; i<21; i++){                            //  0(size) 1-2(addr)   3(xor)  4(Data)                             20(chksum)
        if(i<4){                                            // Keep original 0-4 bytes
          s[0]=*b++;
          s[1]=*b++;
          uint8_t data = strtol((const char*)s, NULL, 16);
          checksum += data;                                 // Compute checksum
          if(i==3)                                          // Byte 3 is the xor value
            xor_byte = data;                                // Use same xor value as original
        }
        else{
          uint8_t data = din[i-4];                          // Get next byte from 128KB binary array
          if(i<20){                                         // index 4 - 19 are data bytes
            data ^= (0x9A + xor_byte);                      // Apply xor
            checksum += data;                               // And update checksum
          }
          else{
            checksum ^= 0xFF;                               // Perform 2's complement on the computed checksum
            checksum += 1;
            data = checksum;                                // index 20 is checksum, apply our computed checksum
          }
          itoa(data, s, 16);                                // convert data to hex string
          convertToUpperCase(s);
          *b++ = s[0];                                      // Replace data in dst file
          *b++ = s[1];
        }
    }
}

void usage(void)
{
  printf("\r\n  tool [ -e | -i ] [input] [output]\r\n\n");
  printf("  This tool extracts/inserts the firmware from/to the Agilent U1253A updater executable.\r\n");
  printf("  Only tested with \"Agilent U1253A Firmware Update V120.exe\"!\r\n\n");
  printf("    -e: Extract the firmware from the updater executable into a file.\r\n");
  printf("        The output will be the original firmware bytes, ready for modifying or disassembling.\r\n\n");
  printf("    -i: Insert the firmware file into an existing updater executable.\r\n\n");
  printf("  Example: tool -e updater.exe firmware.bin (firmware.bin is created)\r\n");
  printf("  Example: tool -i firmware.bin updater.exe (firmware.bin is inserted into updater.exe)\r\n\n");
  printf("  More info at https://www.eevblog.com/forum/microcontrollers/8-bit-78k0-d78f0547-help-with-firmware-modification/\r\n\n");
}



void extract(char * in, char *out){

  printf("Extracting firmware...\r\n");

  outbf = malloc(131072U);                                  // Alloc output buffer (128KB)

  if(outbf == NULL){
    printf("Unable to allocate output buffer!\r\n");
  }

  inbf = file_load(in, &filesize);                          // Read input file

  if(inbf == NULL){
    printf("Unable to read file \"%s\"!\r\n", in);
    terminal_error();
  }

  printf("Read %lu bytes from file \"%s\"\r\n", filesize, in);

  uint8_t * pos=inbf;
  pos = _memmem (pos, filesize, " :100000", 8);             // Search beginning of intel hex structure
  if(pos == NULL){
    printf("Unable to locate firmware start!\r\n");
    terminal_error();
  }

  uint32_t index=0;
  size_t bytes_remaining, bytes_read;

  printf("Found firmware at offset 0x%X\r\n", (uintptr_t)pos - (uintptr_t)inbf);

  while(pos && index<(128U*1024)){
    bytes_read = (uintptr_t)pos - (uintptr_t)inbf;
    bytes_remaining = filesize - bytes_read;
    pos = memchr(pos, ':', bytes_remaining);
    if(pos != NULL){
      pos++;                                                // Skip ":"
      intelHexToArray(pos, &outbf[index]);                  // Convert each intel hex block
      index+=16;                                            // Each block is 16 bytes
    }
  };

  if(index != (128U*1024)){
    printf("Expected 131.072 bytes, but got %lu bytes!\r\n", index);
  }

  if(!file_save(out, outbf, index)){
    printf("Unable to write file \"%s\"!\r\n", out);
    terminal_error();
  }

  free(inbf);
  free(outbf);
  printf("%lu bytes saved to file %s\r\n", index, out);
}


void insert(char * in, char *out){

  printf("Inserting firmware...\r\n");

  inbf = file_load(in, &filesize);                          // Read input file (Firmware, 128KB)

  if(inbf == NULL){
    printf("Unable to from file \"%s\"!\r\n", in);
    terminal_error();
  }
  if(filesize != (128U*1024)){
    printf("Firmware image must be 131.072 bytes, but got %lu bytes!\r\n", filesize);
    terminal_error();
  }
  printf("Read %lu bytes from file \"%s\"\r\n", filesize, in);

  outbf = file_load(out, &filesize);                        // Read output file (Updater exe)
  if(outbf == NULL){
    printf("Unable to read file \"%s\"!\r\n", out);
    terminal_error();
  }
  printf("Read %lu bytes from file \"%s\"\r\n", filesize, out);

  char * pos=outbf;
  pos = _memmem (pos, filesize, " :100000", 8);             // Search beginning of intel hex structure
  if(pos == NULL){
    printf("Unable to locate firmware start!\r\n");
    terminal_error();
  }

  uint32_t index=0;
  size_t bytes_remaining, bytes_read;

  printf("Found firmware at offset 0x%lX\r\n", (uintptr_t)pos - (uintptr_t)outbf);

  while(pos && index<(128U*1024)){
    bytes_read = (uintptr_t)pos - (uintptr_t)outbf;
    bytes_remaining = filesize - bytes_read;
    pos = memchr(pos, ':', bytes_remaining);
    if(pos){
      pos++;                                                // Skip ":"
      arrayToIntelHex(&inbf[index], pos);                   // Create new intel hex blocks from binary input
      index+=16;                                            // Each block is 16 bytes
    }
  };

  if(index != (128U*1024)){
    printf("Expected 131.072 bytes, but got %lu bytes!\r\n", index);
  }

  if(!file_save(out, outbf, filesize)){
    printf("Unable to write file \"%s\"!\r\n", out);
    terminal_error();
  }

  free(inbf);
  free(outbf);
  printf("%lu bytes saved to file \"%s\"\r\n", filesize, out);
}

int main(int argc, char * argv[])
{
  if(argc < 4){
    usage();
    exit(-1);
  }

  if(strcmp(argv[1], "-i") == 0){
    insert(argv[2], argv[3]);
  }
  else if(strcmp(argv[1], "-e") == 0){
    extract(argv[2], argv[3]);
  }
  else{
    printf("Unknown command: \"%s\"\r\n", argv[1]);
    usage();
    exit(-1);
  }

  printf("Done!\r\n");
  return 0;
}
